/* systems.c, systems-specific routines                                 */

/* Copyright (C) 1996, MPEG Software Simulation Group. All Rights Reserved. */

/*
 * Disclaimer of Warranty
 *
 * These software programs are available to the user without any license fee or
 * royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
 * any and all warranties, whether express, implied, or statuary, including any
 * implied warranties or merchantability or of fitness for a particular
 * purpose.  In no event shall the copyright-holder be liable for any
 * incidental, punitive, or consequential damages of any kind whatsoever
 * arising from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and user's
 * customers, employees, agents, transferees, successors, and assigns.
 *
 * The MPEG Software Simulation Group does not represent or warrant that the
 * programs furnished hereunder are free of infringement of any third-party
 * patents.
 *
 * Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
 * are subject to royalty fees to patent holders.  Many of these patents are
 * general enough such that they are unavoidable regardless of implementation
 * design.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "config.h"
#include "global.h"

/* initialize buffer, call once before first getbits or showbits */
/* 
 * parse transport layer as per section 2.4.3.2 of ISO/IEX 13818-1
 * This function only processes PES packets with stream ids 0xE0.
 * It locates the first PES packet that has a stream id of 0xEO and then
 * all subsequent packets that belong to tha stream based on the PID 
 * of the packet. Hence it processes only one Video Elementary Stream.
 */
void Next_PES_Packet()
{
  unsigned int code;

  int transport_indicator;
  int payload_unit_start_indicator;
  int transport_priority;
  int PID;
  int transport_scrambling_control = 0;
  int  adaptation_field = 0;
  int payload_field = 0;
  int continuity_counter = 0;
  int num_bytes_skip = 0;
  static int counter = 0;

  for(;;)
  {
	int PIDa, PIDb;

	ld->Rdptr = ld->Rdmax = ld->Rdbfr + 188;
    code = Get_Byte();

    /* remove system layer byte stuffing */
    while ((code & 0xff) != 0x47) /*Check if sync byte*/
	{
      code = Get_Byte();
	}

	  ld->Rdmax = ld->Rdbfr + 188;

      /* skip pack header (system_clock_reference and mux_rate) */
	  counter++;
	  code = Get_Byte();
	  transport_indicator = code & 0x80 ? 1 : 0;
	  payload_unit_start_indicator = code & 0x40 ? 1 : 0;
	  transport_priority = code & 0x20 ? 1 : 0;
	  PIDa = code & 0x1F;
	  code = Get_Byte();
	  PIDb = code;
	  PID = (PIDa << 8) | PIDb;
	  code = Get_Byte();
	  transport_scrambling_control = (code & 0xc0) >> 6;
	  adaptation_field = code & 0x20 ? 1 : 0;
	  payload_field = code & 0x10 ? 1 : 0;
	  continuity_counter = code & 0x0f;
	  /* skip the adaptation field if present*/
	  if (adaptation_field ==1)
	  {
		num_bytes_skip = Get_Byte(); 
		ld->Rdptr += num_bytes_skip; 
	  }
	  /*if no payload present, skip rest of packet*/
	  if (payload_field == 0)
	  {
		ld->Rdptr = ld->Rdbfr + 188;
		continue;
	  }
	  /*
	   * Skip rest of packet if PID is not equal to the PID of the stream
	   * currently being processed.
	   */
	  if (pesPID != -1 && PID != pesPID)
	  {
		ld->Rdptr = ld->Rdbfr + 188;
		continue;
	  }

	 /*
	  * if start of payload, process the packet only if the stream is a
	  * video elementary stream with id 0xE0 and is the first stream to be
	  * processed
	  */
	 if(payload_unit_start_indicator)
	 {
	   code = Get_Long();
	   /* VIDEO Elementary Streams have code 0x1e0 through 0x1ef */
	   if ( !(code >= 0x1e0 && code <= 0x1ef) && pesPID == -1)
	   {
	  	  ld->Rdptr = ld->Rdbfr + 188;
#ifdef DEBUG
          printf ("%d: Skipping non-VIDEO_ES PES payload with code=0x%04x PID=%d\n", counter, code, PID);
#endif
		  continue;
	   }
	   if (pesPID == -1)
	 	  pesPID = PID;

#ifdef DEBUG	
        printf ("%d: payload_unit_start_indicator: %d\n", counter, payload_unit_start_indicator);
        printf ("%d: PID: 0x%x\n", counter, PID);
        printf ("%d: adaptation_field: %d\n", counter, adaptation_field);
        printf ("%d: payload_field: %d\n", counter, payload_field);
        printf ("%d: adaptation_field_length: %d\n", counter, num_bytes_skip);
#endif
	   if (code == VIDEO_ELEMENTARY_STREAM)
	   {
	     code = Get_Word();
#ifdef DEBUG
	     printf ("%d: #Packet_length = %d\n", counter, code);
#endif
	     code = Get_Byte();
	     if((code>>6)==0x02)
         {
           ld->Rdptr++;
           code=Get_Byte();  /* parse PES_header_data_length */
           ld->Rdptr+=code;    /* advance pointer by PES_header_data_length */
           printf("%d: MPEG-2 PES packet\n", counter);
           return;
         }
         else if(code==0xff)
         {
           /* parse MPEG-1 packet header */
           while((code=Get_Byte())== 0xFF);
         }
         
         /* stuffing bytes */
         if(code>=0x40)
         {
           if(code>=0x80)
           {
             fprintf(stderr,"Error in packet header\n");
             exit(1);
           }
           /* skip STD_buffer_scale */
           ld->Rdptr++;
           code = Get_Byte();
         }
         if(code>=0x30)
         {
           if(code>=0x40)
           {
             fprintf(stderr,"Error in packet header\n");
             exit(1);
           }
           /* skip presentation and decoding time stamps */
           ld->Rdptr += 9;
         }
         else if(code>=0x20)
         {
           /* skip presentation time stamps */
           ld->Rdptr += 4;
         }
         else if(code!=0x0f)
         {
           fprintf(stderr,"Error in packet header\n");
           exit(1);
         }
	   }
	 }
	 /*
	  * If not beginning of payload, process the packet only if the PID of the
	  * PES packet is equal to the PID of the Video Elementary stream already
	  * being processed
	  */
	 if (pesPID == PID)
	 {
	   if (!payload_field)
	   {
	     ld->Rdptr = ld->Rdbfr + 188;
		 continue;
	   }
#ifdef DEBUG
       printf ("%d: payload_unit_start_indicator: %d\n", counter, payload_unit_start_indicator);
       printf ("%d: PID: 0x%x\n", counter, PID);
       printf ("%d: adaptation_field: %d\n", counter, adaptation_field);
       printf ("%d: payload_field: %d\n", counter, payload_field);
       printf ("%d: adaptation_field_length: %d\n", counter, num_bytes_skip);
#endif
	   break;	    
	 }
	 /*fetch next packet in transport stream*/
	 continue;
  }
}

/* initialize buffer, call once before first getbits or showbits */
/* parse system layer, ignore everything we don't need */
void Next_Packet()
{
  unsigned int code;
  int l;

  for(;;)
  {
    code = Get_Long();

    /* remove system layer byte stuffing */
    while ((code & 0xffffff00) != 0x100)
      code = (code<<8) | Get_Byte();

	if ((code & VIDEO_ELEMENTARY_STREAM) == VIDEO_ELEMENTARY_STREAM)
		code = VIDEO_ELEMENTARY_STREAM;
    switch(code)
    {
    case PACK_START_CODE: /* pack header */
      /* skip pack header (system_clock_reference and mux_rate) */
      ld->Rdptr += 8;
      break;
    case VIDEO_ELEMENTARY_STREAM:   
      code = Get_Word();             /* packet_length */
	  ld->Rdmax = ld->Rdptr + code;

      code = Get_Byte();

      if((code>>6)==0x02)
      {
        ld->Rdptr++;
        code=Get_Byte();  /* parse PES_header_data_length */
        ld->Rdptr+=code;    /* advance pointer by PES_header_data_length */
        printf("MPEG-2 PES packet\n");
        return;
      }
      else if(code==0xff)
      {
        /* parse MPEG-1 packet header */
        while((code=Get_Byte())== 0xFF);
      }
       
      /* stuffing bytes */
      if(code>=0x40)
      {
        if(code>=0x80)
        {
          fprintf(stderr,"Error in packet header\n");
          exit(1);
        }
        /* skip STD_buffer_scale */
        ld->Rdptr++;
        code = Get_Byte();
      }

      if(code>=0x30)
      {
        if(code>=0x40)
        {
          fprintf(stderr,"Error in packet header\n");
          exit(1);
        }
        /* skip presentation and decoding time stamps */
        ld->Rdptr += 9;
      }
      else if(code>=0x20)
      {
        /* skip presentation time stamps */
        ld->Rdptr += 4;
      }
      else if(code!=0x0f)
      {
        fprintf(stderr,"Error in packet header\n");
        exit(1);
      }
      return;
    case ISO_END_CODE: /* end */
      /* simulate a buffer full of sequence end codes */
      l = 0;
      while (l<2048)
      {
        ld->Rdbfr[l++] = SEQUENCE_END_CODE>>24;
        ld->Rdbfr[l++] = SEQUENCE_END_CODE>>16;
        ld->Rdbfr[l++] = SEQUENCE_END_CODE>>8;
        ld->Rdbfr[l++] = SEQUENCE_END_CODE&0xff;
      }
      ld->Rdptr = ld->Rdbfr;
      ld->Rdmax = ld->Rdbfr + 2048;
      return;
    default:
      if(code>=SYSTEM_START_CODE)
      {
        /* skip system headers and non-video packets*/
        code = Get_Word();
        ld->Rdptr += code;
      }
      else
      {
        fprintf(stderr,"Unexpected startcode %08x in system layer\n",code);
        exit(1);
      }
      break;
    }
  }
}



void Flush_Buffer32()
{
  int Incnt;
  int offset;

  ld->Bfr = 0;

  Incnt = ld->Incnt;
  Incnt -= 32;

  offset = 2048;

  if (Transport_Stream_Flag && (ld->Rdptr >= ld->Rdmax-4))
  {
    while (Incnt <= 24)
    {
      if (ld->Rdptr >= ld->Rdmax)
        Next_PES_Packet();
      ld->Bfr |= Get_Byte() << (24 - Incnt);
      Incnt += 8;
    }
  }
  else if (System_Stream_Flag && (ld->Rdptr >= ld->Rdmax-4))
  {
    while (Incnt <= 24)
    {
      if (ld->Rdptr >= ld->Rdmax)
        Next_Packet();
      ld->Bfr |= Get_Byte() << (24 - Incnt);
      Incnt += 8;
    }
  }
  else
  {
    while (Incnt <= 24)
    {
      if (ld->Rdptr >= ld->Rdbfr+offset)
        Fill_Buffer();
      ld->Bfr |= *ld->Rdptr++ << (24 - Incnt);
      Incnt += 8;
    }
  }
  ld->Incnt = Incnt;

#ifdef VERIFY 
  ld->Bitcnt += 32;
#endif /* VERIFY */
}


unsigned int Get_Bits32()
{
  unsigned int l;

  l = Show_Bits(32);
  Flush_Buffer32();

  return l;
}


int Get_Long()
{
  int i;

  i = Get_Word();
  return (i<<16) | Get_Word();
}


